from ccip_sdk import CCIPClient
from dotenv import load_dotenv
import os

load_dotenv()

client = CCIPClient(private_key=os.environ.get("PRIVATE_KEY"))

contract = client.deploy_sender_contract("ethereum_sepolia")
print(f"Deployed contract address: {contract}\n")
txn_hash = client.send_tokens_to_sender_contract("ethereum_sepolia", "CCIP-BnM", 0.1)
print(f"Token sent via this transaction hash : {txn_hash}\n")
txn_hash = client.send_eth_to_contract("ethereum_sepolia", 0.05)
print(f"ETH sent via this transaction hash : {txn_hash}\n")
txn_hash = client.allow_destination_chain(current_chain="ethereum_sepolia", destination_chain="arbitrum_sepolia")
print(f"Allowed destination chain arbitrum_sepolia done with txnHash : {txn_hash}\n")

contract = client.deploy_receiver_contract("arbitrum_sepolia")
print(f"Deployed contract address: {contract}\n")
txn_hash = client.allow_source_chain(current_chain="arbitrum_sepolia", sender_chain="ethereum_sepolia")
print(f"Allowed sender chain ethereum_sepolia done with txnHash : {txn_hash}\n")
txn_hash = client.allow_sender_on_receiver(sender_chain="ethereum_sepolia", receiver_chain="arbitrum_sepolia")
print(f"Allowed the sender contract to send messages on reciever chain with txnHash : {txn_hash}\n")

txn_url = client.transfer(sender_chain="ethereum_sepolia", receiver_chain="arbitrum_sepolia", text="Hi dj boi", amount=0.069)
print(f"You can watch the CCIP Transfer here : {txn_url}\n")




from dataclasses import dataclass


@dataclass(frozen=True)
class CCIPConfig:
    """
    config for ccip client
    RPC_URL's for ccip directory based chains 
    """
    chains = {}
    ccip_explorer = "https://ccip.chain.link/#/side-drawer/"



from typing import Optional, Literal
from .utils import (
    get_account_from_private_key, 
    get_ccip_directory_data, 
    deploy_contract, 
    send_erc20_to_contract,
    send_native_eth,
    allowlistDestinationChain,
    allowlistSourceChain,
    allowlistSender,
    transfer,
    getLastReceivedMessage,
    withdraw_token,
    withdraw_eth
)
from pathlib import Path
import time

class CCIPClient:
    """
    CCIPClient handles CCIP state logic for the SDK.
    """
    def __init__(self, **kwargs: str) -> None:
        """
        Initializes the CCIPClient with a private key.

        :param private_key: The private key for the client.
        """
        for key, value in kwargs.items():
            if key == "private_key":
                private_key = value
            else:
                raise ValueError(f"Unknown argument {key}.")
        self.private_key = private_key
        self.account = get_account_from_private_key(private_key)
        self.sender_contracts = {}
        self.receiver_contracts = {}
        SCRIPT_DIR = Path(__file__).parent.absolute()
        DATA_FILE = SCRIPT_DIR / "ccip_directory" / "data.json"
        CONTRACT_FILE = SCRIPT_DIR / "contracts" / "artifacts" / "contracts" / "CCIPContract.sol" / "CCIPContract.json"
        self.contract_compiled_file = str(CONTRACT_FILE)
        self.ccip_directory_path = str(DATA_FILE)
        self.chains_data = get_ccip_directory_data(self.ccip_directory_path)
        self.chains = list(self.chains_data.keys())
        self.tokens = list(self.chains_data["ethereum_sepolia"]["tokens"].keys())


    def deploy_sender_contract(self, chain: str) -> str:
        """
        deploy the sender contract and store address

        :param chain: The name of the chain
        """
        self.validate_chain(chain)
        
        rpc = self.chains_data[chain]["rpc"]
        router = self.chains_data[chain]["router"]
        link = self.chains_data[chain]["tokens"]["LINK"]
        print("Deploying contract on chain:", chain)
        contract_address = deploy_contract(
            rpc=rpc,
            router=router,
            link=link,
            filepath=self.contract_compiled_file,
            account=self.account
        )
        
        self.sender_contracts[chain] = contract_address
        return contract_address
    
    def deploy_receiver_contract(self, chain: str) -> str:
        """
        deploy the receiver contract and store address

        :param chain: The name of the chain
        """
        self.validate_chain(chain)
        
        rpc = self.chains_data[chain]["rpc"]
        router = self.chains_data[chain]["router"]
        link = self.chains_data[chain]["tokens"]["LINK"]
        print("Deploying contract on chain:", chain)
        contract_address = deploy_contract(
            rpc=rpc,
            router=router,
            link=link,
            filepath=self.contract_compiled_file,
            account=self.account
        )
        
        self.receiver_contracts[chain] = contract_address
        return contract_address

    def send_tokens_to_sender_contract(self, chain: str, token: str, amount: float) -> str:
        """
        send the given token to the contract
        :param chain: the network on which the contract is 
        :param token: the token which is to be sent to the contract
        :param amount: the amount to be sent in normal form
        """
        time.sleep(2)
        self.validate_chain(chain)
        self.validate_token(token)

        token_address = self.chains_data[chain]["tokens"].get(token)
        rpc = self.chains_data[chain]["rpc"]

        if not token_address:
            raise ValueError(f"Token {token} not supported on chain {chain}")
        contract_address = self.sender_contracts.get(chain)
        if not contract_address:
            raise ValueError(f"Contract is not deployed on {chain}")
        print(f"Sending {token} to contract on chain : {chain}")
        txn_hash = send_erc20_to_contract(rpc, token_address, self.account, amount, contract_address)
        return txn_hash
    
    def send_eth_to_contract(self, chain:str, amount: float) -> str:
        time.sleep(3)
        self.validate_chain(chain)
        rpc = self.chains_data[chain]["rpc"]

        contract_address = self.sender_contracts.get(chain)
        if not contract_address:
            raise ValueError(f"Contract is not deployed on {chain}")
        print(f"Sending native eth to contract on chain : {chain}")
        txn_hash = send_native_eth(rpc, self.account, amount, contract_address)
        return txn_hash
    
    def allow_destination_chain(self, current_chain = None, destination_chain = None) -> str:
        time.sleep(3)
        if not current_chain or not destination_chain: 
            raise ValueError(f"Please provide valid current and destination chain")
        if not self.validate_chain(current_chain) or not self.sender_contracts.get(current_chain):
            raise ValueError("Chain is not deployed or not a part of CCIP directory")
        rpc = self.chains_data[current_chain]["rpc"]
        if not self.chains_data.get(destination_chain):
            raise ValueError(f"Destination chain not found in CCIP Directory")
        
        chainSelector = self.chains_data[destination_chain]["chain_selector"]
        contract = self.sender_contracts[current_chain]
        if not contract:
            raise ValueError(f"No Sender contract found on {current_chain}")
        print(f"Allowing the destination chain from current chain : {current_chain}")
        txn_hash = allowlistDestinationChain(rpc, chainSelector, contract, self.account, self.contract_compiled_file)
        return txn_hash
    
    def allow_source_chain(self, current_chain = None, sender_chain = None) -> str:
        time.sleep(3)
        if not current_chain or not sender_chain: 
            raise ValueError(f"Please provide valid current and destination chain")
        if not self.validate_chain(current_chain) or not self.sender_contracts.get(sender_chain):
            raise ValueError("Chain is not deployed or not a part of CCIP directory")
        if not self.receiver_contracts.get(current_chain):
            raise ValueError(f"Contract not deployed on chain : {current_chain}")
        rpc = self.chains_data[current_chain]["rpc"]
        if not self.chains_data.get(sender_chain):
            raise ValueError(f"Destination chain not found in CCIP Directory")
        
        chainSelector = self.chains_data[sender_chain]["chain_selector"]
        contract = self.receiver_contracts[current_chain]
        if not contract:
            raise ValueError(f"No Sender contract found on {current_chain}")
        print(f"Allowing the sender chain from receiver chain : {current_chain}")
        txn_hash = allowlistSourceChain(rpc, chainSelector, contract, self.account, self.contract_compiled_file)
        return txn_hash
    
    def allow_sender_on_receiver(self, sender_chain=None, receiver_chain=None):
        time.sleep(3)
        if not sender_chain or not receiver_chain: 
            raise ValueError(f"Please provide valid current and destination chain")
        if not self.sender_contracts.get(sender_chain) or not self.receiver_contracts.get(receiver_chain):
            raise ValueError("Chain is not deployed or not a part of CCIP directory")
        rpc = self.chains_data[receiver_chain]["rpc"]
        sender_contract = self.sender_contracts[sender_chain]
        receiver_contract = self.receiver_contracts[receiver_chain]
        print(f"Allowing the sender contract : {sender_contract} on reciever contract : {receiver_contract}")
        txn_hash = allowlistSender(rpc, sender_contract, receiver_contract, self.account, self.contract_compiled_file)
        return txn_hash
    
    def transfer(self, sender_chain=None, receiver_chain=None, text="None", amount=0.1, token="CCIP-BnM"):
        time.sleep(3)
        if not sender_chain or not receiver_chain: 
            raise ValueError(f"Please provide valid current and destination chain")
        if not self.sender_contracts.get(sender_chain) or not self.receiver_contracts.get(receiver_chain):
            raise ValueError("Chain is not deployed or not a part of CCIP directory")
        rpc = self.chains_data[sender_chain]["rpc"]
        sender_contract = self.sender_contracts[sender_chain]
        receiver_contract = self.receiver_contracts[receiver_chain]
        destination_chain_selector = self.chains_data[receiver_chain]["chain_selector"]
        token_address = self.chains_data[sender_chain]["tokens"][token]
        print(f"CCIP Transfer from chain {sender_chain} to destination chain {receiver_chain}")
        txn_hash = transfer(rpc, sender_contract, destination_chain_selector, receiver_contract, text, token_address, amount, self.account, self.contract_compiled_file)
        return f"https://ccip.chain.link/tx/{txn_hash}"
    
    def get_message_on_reciever_contract(self, receiver_chain: str, contract = None) -> str:
        """
        Function to get the message on the receiver contract
        :param receiver_chain: The chain on which the receiver contract is deployed
        :param[Optional] contract : contract address on the receiver network
        """
        if receiver_chain not in self.receiver_contracts:
            raise ValueError(f"Receiver contract not deployed on {receiver_chain}")
        
        rpc = self.chains_data[receiver_chain]["rpc"]
        if contract is None:
            contract_address = self.receiver_contracts[receiver_chain]
        
        print(f"Getting last received message on receiver contract on chain : {receiver_chain}")
        message = getLastReceivedMessage(rpc, contract_address, self.contract_compiled_file)
        return message
    
    def withdraw_token_to_wallet(self, chain: str, token: str, beneficiary: str) -> str:
        """
        :param chain: The chain on which the contract is deployed
        :param beneficiary: The address to which the tokens will be withdrawn
        """
        if chain not in self.chains:
            raise ValueError(f"Sender contract not deployed on {chain}")
        
        rpc = self.chains_data[chain]["rpc"]
        contract_address = self.receiver_contracts[chain]
        token_address = self.chains_data[chain]["tokens"][token]
        
        print(f"Withdrawing tokens to wallet on chain : {chain}")
        txn_hash = withdraw_token(rpc, contract_address, token_address, beneficiary, self.account, self.contract_compiled_file)
        return txn_hash
    
    def withdraw_eth_to_wallet(self, chain: str, beneficiary: str) -> str:
        """
        :param chain: The chain on which the contract is deployed
        :param beneficiary: The address to which the ETH will be withdrawn
        """
        if chain not in self.chains:
            raise ValueError(f"Sender contract not deployed on {chain}")
        
        rpc = self.chains_data[chain]["rpc"]
        contract_address = self.receiver_contracts[chain]
        
        print(f"Withdrawing ETH to wallet on chain : {chain}")
        txn_hash = withdraw_eth(rpc, contract_address, beneficiary, self.account, self.contract_compiled_file)
        return txn_hash

    def validate_chain(self, chain: str):
        """
        Function that checks if the chain is valid
        """
        if chain not in self.chains:
            raise ValueError(f"Chain {chain} is not supported.")
        return True
    
    def validate_token(self, token: str):
        """
        Function that checks if the token is valid and supported
        """
        if token not in self.tokens:
            raise ValueError(f"Token {token} is not supported")
        return True

    


import json
from eth_account import Account
from web3 import Web3
import importlib.resources

multiplier = 2

ERC20_ABI = [
        {
            "constant": False,
            "inputs": [
                {"name": "_to", "type": "address"},
                {"name": "_value", "type": "uint256"}
            ],
            "name": "transfer",
            "outputs": [{"name": "", "type": "bool"}],
            "type": "function"
        }
]

def get_account_from_private_key(private_key: str):
    """
    Returns an account object from a private key.

    :param private_key: The private key to create the account from.
    :return: An account object.
    """
    return Account.from_key(private_key)

def get_contract_interface(file_path: str) -> dict:
    """
    Reads a contract ABI from a JSON file.

    :param file_path: The path to the JSON file containing the ABI.
    :return: The ABI as a dictionary.
    """
    
    with importlib.resources.open_text(
        "ccip_sdk.contracts.artifacts.contracts.CCIPContract", 
        "CCIPContract.json"
    ) as f:
        return json.load(f)
    
def get_ccip_directory_data(file_path: str) -> dict:
    """
    Reads CCIP directory data from a JSON file.

    :param file_path: The path to the JSON file containing the CCIP directory data.
    :return: The CCIP directory data as a dictionary.
    """
    with importlib.resources.open_text("ccip_sdk.ccip_directory", "data.json") as f:
        return json.load(f)
    
def deploy_contract(rpc: str, router: str, link: str, filepath: str, account: Account, multiplier=multiplier):
    """
    Deploy contract on the given chain

    :param rpc: The rpc url of the chain 
    :param router: The address of the CCIP router
    :param link: The address of the LINK token
    :param filepath: The path to the interface containing abi, bytecode
    :return : The contract address in string format
    """
    w3 = Web3(Web3.HTTPProvider(rpc))
    contract_interface = get_contract_interface(filepath)
    abi = contract_interface["abi"]
    bytecode = contract_interface["bytecode"]

    nonce = max(
        w3.eth.get_transaction_count(account.address, 'pending'),
        w3.eth.get_transaction_count(account.address, 'latest')
    )
    contract = w3.eth.contract(abi=abi, bytecode=bytecode)

    balance = w3.eth.get_balance(account.address)
    print(f"Account balance: {w3.from_wei(balance, 'ether')} ETH")

    gas_estimate = contract.constructor(router, link).estimate_gas({"from": account.address})

    transaction = {
        'nonce': nonce,
        'gas': int(gas_estimate * multiplier),
        'gasPrice': w3.eth.gas_price,
        'from': account.address,
    }
    try:
        # Deploy contract
        contract_tx = contract.constructor(router, link).build_transaction(transaction)
        
        signed_tx = w3.eth.account.sign_transaction(contract_tx, account.key.hex())
        tx_hash = w3.eth.send_raw_transaction(signed_tx.raw_transaction)
        
        # Wait for transaction receipt
        tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
        contract_address = tx_receipt['contractAddress']
        
        print(f"Contract deployed successfully!")
        print(f"Contract address: {contract_address}")
        print(f"Transaction hash: {tx_hash.hex()}")
        
        return contract_address
        
    except Exception as e:
        print(f"Error deploying contract: {str(e)}")
        return None

def send_erc20_to_contract(rpc: str, token: str, account: Account, amount: float, contract: str, multiplier=multiplier):
    w3 = Web3(Web3.HTTPProvider(rpc))
    nonce = max(
        w3.eth.get_transaction_count(account.address, 'pending'),
        w3.eth.get_transaction_count(account.address, 'latest')
    )
    amount = Web3.to_wei(amount, "ether")
    token_contract = w3.eth.contract(address=Web3.to_checksum_address(token), abi=ERC20_ABI)
    gas_estimate = token_contract.functions.transfer(
        Web3.to_checksum_address(contract),
        amount
    ).estimate_gas({"from": account.address})
    
    try:

        tx = token_contract.functions.transfer(
            Web3.to_checksum_address(contract),
            amount
        ).build_transaction({
            "chainId": w3.eth.chain_id,
            "gas": int(gas_estimate * multiplier),
            "gasPrice": w3.eth.gas_price,
            "nonce": nonce,
        })

        signed_tx = w3.eth.account.sign_transaction(tx, account.key.hex())
        tx_hash = w3.eth.send_raw_transaction(signed_tx.raw_transaction)

        print(f"Transaction sent! Hash: {tx_hash.hex()}")
        return f"0x{tx_hash.hex()}"
    except Exception as e:
        print(f"Error sending token to contract: {str(e)}")
        return None
    
def send_native_eth(rpc: str, account: Account, amount: float, contract: str, multiplier=multiplier):
    w3 = Web3(Web3.HTTPProvider(rpc))
    nonce = max(
        w3.eth.get_transaction_count(account.address, 'pending'),
        w3.eth.get_transaction_count(account.address, 'latest')
    )
    amount = Web3.to_wei(amount, "ether")
    tx = {
        'to': Web3.to_checksum_address(contract),
        'value': amount,
        'gas': 25000,
        'gasPrice': w3.eth.gas_price,
        'nonce': nonce,
        'chainId': w3.eth.chain_id,
    }
    try:
        signed_tx = w3.eth.account.sign_transaction(tx, private_key=account.key.hex())
        tx_hash = w3.eth.send_raw_transaction(signed_tx.raw_transaction)
        tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
        print(f"Transaction hash: 0x{tx_hash.hex()}")
        return f"0x{tx_hash.hex()}"
    except Exception as e:
        print(f"Error transferring native tokens: {str(e)}")
        return False

def allowlistDestinationChain(rpc: str, chainSelector: int, contract: str, account: Account, filepath: str, multiplier=multiplier):
    w3 = Web3(Web3.HTTPProvider(rpc))
    nonce = max(
        w3.eth.get_transaction_count(account.address, 'pending'),
        w3.eth.get_transaction_count(account.address, 'latest')
    )
    contract_interface = get_contract_interface(filepath)
    abi = contract_interface["abi"]
    contract = w3.eth.contract(address=contract, abi=abi)
    chainSelector = int(chainSelector)
    allowed = True
    gas_estimate = contract.functions.allowlistDestinationChain(chainSelector, allowed).estimate_gas({"from": account.address})

    transaction = {
        'nonce': nonce,
        'gas': int(gas_estimate * multiplier),
        'gasPrice': w3.eth.gas_price,
        'from': account.address,
    }

    try:
        tx = contract.functions.allowlistDestinationChain(chainSelector, allowed).build_transaction(transaction)
        signed_tx = w3.eth.account.sign_transaction(tx, account.key.hex())
        tx_hash = w3.eth.send_raw_transaction(signed_tx.raw_transaction)    
        tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
        print(f"Transaction hash: 0x{tx_hash.hex()}")
        return f"0x{tx_hash.hex()}"
    except Exception as e:
        print(f"Error allowing destination chain: {str(e)}")
        return False

def allowlistSourceChain(rpc: str, chainSelector: int, contract: str, account: Account, filepath: str, multiplier=multiplier):
    w3 = Web3(Web3.HTTPProvider(rpc))
    nonce = max(
        w3.eth.get_transaction_count(account.address, 'pending'),
        w3.eth.get_transaction_count(account.address, 'latest')
    )
    contract_interface = get_contract_interface(filepath)
    abi = contract_interface["abi"]
    contract = w3.eth.contract(address=contract, abi=abi)
    allowed = True
    chainSelector = int(chainSelector)
    gas_estimate = contract.functions.allowlistSourceChain(chainSelector, allowed).estimate_gas({"from": account.address})

    transaction = {
        'nonce': nonce,
        'gas': int(gas_estimate * multiplier),
        'gasPrice': w3.eth.gas_price,
        'from': account.address,
    }

    try:
        tx = contract.functions.allowlistSourceChain(chainSelector, allowed).build_transaction(transaction)
        signed_tx = w3.eth.account.sign_transaction(tx, account.key.hex())
        tx_hash = w3.eth.send_raw_transaction(signed_tx.raw_transaction)    
        tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
        print(f"Transaction hash: 0x{tx_hash.hex()}")
        return f"0x{tx_hash.hex()}"
    except Exception as e:
        print(f"Error allowing destination chain: {str(e)}")
        return False

def allowlistSender(rpc: str, sender_contract: str, current_contract: str, account: Account, filepath: str, multiplier=multiplier):
    w3 = Web3(Web3.HTTPProvider(rpc))
    nonce = max(
        w3.eth.get_transaction_count(account.address, 'pending'),
        w3.eth.get_transaction_count(account.address, 'latest')
    )
    contract_interface = get_contract_interface(filepath)
    abi = contract_interface["abi"]
    contract = w3.eth.contract(address=current_contract, abi=abi)

    allowed = True
    gas_estimate = contract.functions.allowlistSender(sender_contract, allowed).estimate_gas({"from": account.address})

    transaction = {
        'nonce': nonce,
        'gas': int(gas_estimate * multiplier),
        'gasPrice': w3.eth.gas_price,
        'from': account.address,
    }

    try:
        tx = contract.functions.allowlistSender(sender_contract, allowed).build_transaction(transaction)
        signed_tx = w3.eth.account.sign_transaction(tx, account.key.hex())
        tx_hash = w3.eth.send_raw_transaction(signed_tx.raw_transaction)    
        tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
        print(f"Transaction hash: 0x{tx_hash.hex()}")
        return f"0x{tx_hash.hex()}"
    except Exception as e:
        print(f"Error allowing destination chain: {str(e)}")
        return False
    
def transfer(rpc: str, current_contract: str, destination_chain_selector: int, receiver_contract: str, text: str, token: str, amount: float, account: Account, filepath: str, multiplier=multiplier):
    w3 = Web3(Web3.HTTPProvider(rpc))
    nonce = max(
        w3.eth.get_transaction_count(account.address, 'pending'),
        w3.eth.get_transaction_count(account.address, 'latest')
    )
    contract_interface = get_contract_interface(filepath)
    abi = contract_interface["abi"]
    contract = w3.eth.contract(address=current_contract, abi=abi)
    chainSelector = int(destination_chain_selector)
    amount = Web3.to_wei(amount, "ether")
    gas_estimate = contract.functions.sendMessagePayNative(chainSelector, receiver_contract, text, token, amount).estimate_gas({"from": account.address})

    transaction = {
        'nonce': nonce,
        'gas': int(gas_estimate * multiplier),
        'gasPrice': w3.eth.gas_price,
        'from': account.address,
    }

    try:
        tx = contract.functions.sendMessagePayNative(chainSelector, receiver_contract, text, token, amount).build_transaction(transaction)
        signed_tx = w3.eth.account.sign_transaction(tx, account.key.hex())
        tx_hash = w3.eth.send_raw_transaction(signed_tx.raw_transaction)    
        tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
        print(f"Transaction hash: 0x{tx_hash.hex()}")
        return f"0x{tx_hash.hex()}"
    except Exception as e:
        print(f"Error allowing destination chain: {str(e)}")
        return False
    
def getLastReceivedMessage(rpc: str, contract: str, filepath: str):
    w3 = Web3(Web3.HTTPProvider(rpc))
    contract_interface = get_contract_interface(filepath)
    abi = contract_interface["abi"]
    contract = w3.eth.contract(address=contract, abi=abi)
    
    try:
        last_message = contract.functions.getLastReceivedMessage().call()
        print(f"Last received message: {last_message}")
        return last_message
    except Exception as e:
        print(f"Error getting last received message: {str(e)}")
        return None

def withdraw_token(rpc: str, contract: str, token: str, beneficiary: str, account: Account, filepath: str, multiplier=multiplier):
    w3 = Web3(Web3.HTTPProvider(rpc))
    nonce = max(
        w3.eth.get_transaction_count(account.address, 'pending'),
        w3.eth.get_transaction_count(account.address, 'latest')
    )
    contract_interface = get_contract_interface(filepath)
    abi = contract_interface["abi"]
    contract = w3.eth.contract(address=contract, abi=abi)

    token = Web3.to_checksum_address(token)
    beneficiary = Web3.to_checksum_address(beneficiary)
    gas_estimate = contract.functions.withdrawToken(token, beneficiary).estimate_gas({"from": account.address})

    transaction = {
        'nonce': nonce,
        'gas': int(gas_estimate * multiplier),
        'gasPrice': w3.eth.gas_price,
        'from': account.address,
    }

    try:
        tx = contract.functions.withdrawToken(token, beneficiary).build_transaction(transaction)
        signed_tx = w3.eth.account.sign_transaction(tx, account.key.hex())
        tx_hash = w3.eth.send_raw_transaction(signed_tx.raw_transaction)    
        tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
        print(f"Transaction hash: 0x{tx_hash.hex()}")
        return f"0x{tx_hash.hex()}"
    except Exception as e:
        print(f"Error withdrawing token: {str(e)}")
        return False
    
def withdraw_eth(rpc: str, contract: str, beneficiary: str, account: Account, filepath: str, multiplier=multiplier):
    w3 = Web3(Web3.HTTPProvider(rpc))
    nonce = max(
        w3.eth.get_transaction_count(account.address, 'pending'),
        w3.eth.get_transaction_count(account.address, 'latest')
    )
    contract_interface = get_contract_interface(filepath)
    abi = contract_interface["abi"]
    contract = w3.eth.contract(address=contract, abi=abi)

    token = Web3.to_checksum_address(token)
    beneficiary = Web3.to_checksum_address(beneficiary)
    gas_estimate = contract.functions.withdraw(beneficiary).estimate_gas({"from": account.address})

    transaction = {
        'nonce': nonce,
        'gas': int(gas_estimate * multiplier),
        'gasPrice': w3.eth.gas_price,
        'from': account.address,
    }

    try:
        tx = contract.functions.withdraw(beneficiary).build_transaction(transaction)
        signed_tx = w3.eth.account.sign_transaction(tx, account.key.hex())
        tx_hash = w3.eth.send_raw_transaction(signed_tx.raw_transaction)    
        tx_receipt = w3.eth.wait_for_transaction_receipt(tx_hash)
        print(f"Transaction hash: 0x{tx_hash.hex()}")
        return f"0x{tx_hash.hex()}"
    except Exception as e:
        print(f"Error withdrawing token: {str(e)}")
        return False


from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
readme_path = this_directory / "README.md"

# Read README with encoding and fallback
try:
    long_description = readme_path.read_text(encoding="utf-8")
except FileNotFoundError:
    long_description = ""

setup(
    name="ccip-sdk",
    version="0.1.125",
    description="The simplest Python SDK for Chainlink CCIP that turns complex cross-chain operations into 15 lines of code.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Dhananjay Pai",
    author_email="dhananjay2002pai@gmail.com",
    url="https://github.com/dhananjaypai08/ccip-sdk",
    license="MIT",
    packages=find_packages(include=["ccip_sdk", "ccip_sdk.*"]),
    include_package_data=True,
    package_data={
        "ccip_sdk": [
            "ccip_directory/*.json",
            "ccip_directory/**/*.json",
            "contracts/artifacts/contracts/CCIPContract/*.json",
        ],
    },
    install_requires=[
        "web3",
        "eth_account",
        "requests",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Developers",
        "Operating System :: OS Independent",
    ],
    python_requires=">=3.8",
)